"""Enum classes for the GeneWeaver project."""
from enum import Enum


class CurationAssignment(int, Enum):
    """Enum for the different types of curation assignments."""

    UNASSIGNED = 1
    ASSIGNED = 2
    READY_FOR_REVIEW = 3
    REVIEWED = 4
    APPROVED = 5


class GenesetScoreTypeStr(str, Enum):
    """Enum for the different types of geneset scores."""

    P_VALUE = "p-value"
    Q_VALUE = "q-value"
    BINARY = "binary"
    CORRELATION = "correlation"
    EFFECT = "effect"


class GenesetScoreType(int, Enum):
    """Integer based Enum for the different types of geneset scores."""

    P_VALUE = 1
    Q_VALUE = 2
    BINARY = 3
    CORRELATION = 4
    EFFECT = 5


class GenesetAccess(str, Enum):
    """Enum for the different types of geneset access."""

    PRIVATE = "private"
    PUBLIC = "public"


class AnnotationType(str, Enum):
    """Enum for the different types of annotations."""

    MONARCH = "monarch"
    NCBO = "ncbo"


class AdminLevel(int, Enum):
    """Enum for the different levels of admin access."""

    NORMAL_USER = 0
    CURATOR = 1
    ADMIN = 2
    ADMIN_WITH_DEBUG = 3



"""Service code for interacting with the PubMed API.

The main entrypoint is the `get_publication` function, which takes a PubMed ID
and returns a `PublicationInfo` object.

Top-level functions:
    get_publication: Get publication info from PubMed API

PubMed API Function:
    get_xml_for_pubmed_id: Get XML for a PubMed ID from the PubMed API

XML Parsing Functions:
    extract_fields: Extract all the required fields from PubMed XML needed to initialize
    a PublicationInfo object into a dictionary, combination of the following functions:
        - extract_top_level_fields: Extract top-level fields from PubMed XML
        - extract_date: Extract date field from PubMed XML
        - extract_authors: Extract authors field from PubMed XML

XML Utility Functions:
    format_author_node: Format an author node from PubMed XML
    authors_are_complete: Check if authors are complete according PubMed XML
"""
from xml.etree import ElementTree

import requests
from geneweaver.core.config import settings
from geneweaver.core.exc import ExternalAPIError
from geneweaver.core.schema.publication import PublicationInfo
from geneweaver.core.utils import add_to_dict_if_not_none

PUBMED_SVC_URL = settings.SERVICE_URLS.PUBMED_XLM_SVC_URL


def get_publication(pubmed_id: str) -> PublicationInfo:
    """Get publication info from PubMed API.

    :param pubmed_id: The PubMed ID

    :returns: The publication info
    """
    publication_xml = get_xml_for_pubmed_id(pubmed_id)

    publication_fields = extract_fields(publication_xml)

    return PublicationInfo(**publication_fields)


def get_xml_for_pubmed_id(pubmed_id: str) -> ElementTree.XML:
    """Get XML for a PubMed ID from the PubMed API.

    :param pubmed_id: The PubMed ID

    :returns: The XML for the PubMed ID
    """
    response = requests.get(PUBMED_SVC_URL.format(pubmed_id))

    if not response.ok:
        raise ExternalAPIError(
            f"Error retrieving publication info from PubMed API: {response.status_code}"
        )

    return ElementTree.fromstring(response.text.encode("utf-8"))


def extract_fields(publication_xml: ElementTree.XML) -> dict:
    """Extract publication fields from XML.

    :param publication_xml: The XML to extract fields from

    :returns: The extracted fields
    """
    publication_fields = extract_top_level_fields(publication_xml)

    publication_fields.update(extract_date(publication_xml))

    publication_fields.update(extract_authors(publication_xml))

    return publication_fields


TOP_LEVEL_XML_FIELD_MAPS = {
    "title": ".//ArticleTitle",
    "abstract": ".//AbstractText",
    "journal": ".//Journal/Title",
    "volume": ".//Volume",
    "pages": ".//MedlinePgn",
}


def extract_top_level_fields(publication_xml: ElementTree.XML) -> dict:
    """Extract publication top level fields from XML.

    :param publication_xml: The XML to extract top level fields from

    :returns: The extracted top level fields
    """
    publication_fields = dict()

    for field, xpath in TOP_LEVEL_XML_FIELD_MAPS.items():
        value = publication_xml.findtext(xpath)
        add_to_dict_if_not_none(publication_fields, field, value)

    return publication_fields


def extract_date(publication_xml: ElementTree.XML) -> dict:
    """Extract publication date from XML.

    :param publication_xml: The XML to extract the date from

    :returns: The extracted date
    """
    publication_date = dict()

    pub_date_node = publication_xml.find(".//PubDate")

    if pub_date_node is not None:
        pub_date = pub_date_node.findtext("Year")
        add_to_dict_if_not_none(publication_date, "year", pub_date)

        pub_date = pub_date_node.findtext("Month")
        add_to_dict_if_not_none(publication_date, "month", pub_date)

        pub_date = pub_date_node.findtext("Day")
        add_to_dict_if_not_none(publication_date, "day", pub_date)

    return publication_date


def format_author_node(author_node: ElementTree.XML) -> str:
    """Format author node to string.

    :param author_node: The author node to format

    :returns: The formatted author node as a string.
    """
    name_parts = []

    fore_name = author_node.findtext("ForeName")

    if fore_name:
        name_parts.append(fore_name)
    else:
        initials = author_node.findtext("Initials")

        if initials:
            name_parts.append(initials)

    last_name = author_node.findtext("LastName")

    if last_name:
        name_parts.append(last_name)

    return " ".join(name_parts)


def extract_authors(publication_xml: ElementTree.XML) -> dict:
    """Extract publication authors from XML.

    :param publication_xml: The XML to extract authors from

    :returns: The extracted authors
    """
    publication_authors = dict()

    author_list_node = publication_xml.find(".//AuthorList")

    if author_list_node:
        author_nodes = author_list_node.findall(".//Author")
        authors = [format_author_node(author_node) for author_node in author_nodes]

        if not authors_are_complete(author_list_node):
            authors.append("et al.")

        add_to_dict_if_not_none(publication_authors, "authors", authors)

    return publication_authors


def authors_are_complete(author_list_node: ElementTree.XML) -> bool:
    """Check if authors are complete.

    Authors are considered completed unless the CompleteYN attribute is present and set
    to 'N'.

    :param author_list_node: The author list node to check

    :returns: Whether the authors are complete
    """
    complete = True

    try:
        # If the CompleteYN attribute is present and set to 'N', the author list is not
        # complete. Documentation recommends taking action when this is the case, but
        # does not specify taking action in any other case.
        # https://www.nlm.nih.gov/bsd/licensee/elements_descriptions.html
        if author_list_node.attrib["CompleteYN"] == "N":
            complete = False
    except KeyError:
        pass

    return complete



"""Group schemas."""
import datetime
from typing import List

from geneweaver.core.schema.stubgenerator import StubGenerator
from pydantic import BaseModel


class Group(BaseModel):
    """Group schema."""

    id: int  # noqa: A003
    name: str
    private: bool
    created: datetime.date
    stubgenerators: List[StubGenerator]


class UserAdminGroup(BaseModel):
    """User admin group schema."""

    name: str
    public: bool
    created: datetime.date



"""Stub generator schemas."""
from pydantic import BaseModel


class StubGenerator(BaseModel):
    """Stub generator schema."""

    id: int  # noqa: A003
    name: str
    querystring: str
    last_update: str



"""Utility functions for the parser module."""
from pathlib import Path
from typing import Dict, List

from geneweaver.core.parse.enum import FileType
from geneweaver.core.types import StringOrPath


def get_file_type(file_path: StringOrPath) -> FileType:
    """Determine if a file at a given path is a csv or xlsx file.

    :param file_path: Path to the file.

    :returns: The file type.
                - 'csv' if the file is a CSV file,
                - 'xlsx' if the file is an Excel file.

    :raises ValueError: If the file type is not supported.
    """
    file_path = Path(file_path)  # convert to Path object if not already
    extension = file_path.suffix.lower()

    if extension == ".csv":
        return FileType.CSV
    elif extension == ".xlsx":
        return FileType.EXCEL
    elif extension == ".txt":
        return FileType.TEXT
    else:
        raise ValueError(f"Unsupported file type {extension}.")


def read_file_content(file_path: StringOrPath) -> str:
    """Read the content of the file at a given path.

    :param file_path: Path to the file.

    :returns: The content of the file as a string.

    :raises FileNotFoundError: If the file does not exist.
    :raises IsADirectoryError: If the path points to a directory.
    :raises PermissionError: If the file is not readable.
    """
    with open(file_path, "r") as f:
        content = f.read()

    return content


def replace_keys(
    data: List[Dict[str, str]], new_keys: List[str]
) -> List[Dict[str, str]]:
    """Replace keys in a list of dictionaries.

    :param data: Original list of dictionaries.
    :param new_keys: List of new keys. The order should correspond to the order of
    original keys.

    :returns: A new list of dictionaries with replaced keys.
    """
    new_data = []
    for row in data:
        new_row = dict(zip(new_keys, row.values()))  # noqa: B905
        new_data.append(new_row)
    return new_data



"""Exceptions related to file parsing."""


class UnsupportedFileTypeError(Exception):
    """Custom exception for when a file type is not supported."""

    pass


class EmptyFileError(UnsupportedFileTypeError):
    """Custom exception for when a file is empty.

    Attributes
    ----------
        file_path -- the path of the file that is empty
        message -- explanation of the error
    """

    def __init__(
        self: "EmptyFileError", file_path: str, message: str = "File is empty."
    ) -> None:
        """Initialize the exception."""
        self.file_path = file_path
        self.message = message
        super().__init__(self.message)

    def __str__(self: "EmptyFileError") -> str:
        """Return a string representation of the exception."""
        return f"{self.file_path} -> {self.message}"


class NotAHeaderRowError(Exception):
    """Raised when a row is not a header row."""

    pass


class InvalidBatchValueLineError(Exception):
    """Raised when a value line is invalid."""

    pass


class MultiLineStringError(Exception):
    """Raised when a string is multiline (but shouldn't be)."""

    pass


class IgnoreLineError(Exception):
    """Raised when a line should be ignored."""

    pass


class MissingRequiredHeaderError(Exception):
    """Raised when a required header is missing."""

    pass


class InvalidScoreThresholdError(Exception):
    """Raised when a score threshold is invalid."""

    pass



"""Project schemas."""
from pydantic import BaseModel


class Project(BaseModel):
    """Project schema."""

    id: int  # noqa: A003
    name: str
    groups: list
    session_id: str
    created: str
    notes: str
    star: str


class ProjectCreate(BaseModel):
    """Project create schema."""

    name: str
    notes: str



"""Gene schema."""
import datetime
from typing import List, Optional, Any

from pydantic import BaseModel, Field


class Gene(BaseModel):
    """Gene schema."""

    id: int  # noqa: A003
    reference_id: str
    gene_database: str
    species: str
    preferred: bool
    date: str


class GeneRow(BaseModel):
    """Gene schema for database row."""

    ode_gene_id: int
    ode_ref_id: str
    gdb_id: int
    sp_id: int
    ode_pref: bool
    ode_date: Optional[str]
    old_ode_gene_ids: Optional[List[int]]


class GeneValue(BaseModel):
    """A gene value."""

    symbol: str
    value: float

    class Config:
        """Pydantic config."""

        allow_mutation = False

    def __hash__(self: "GeneValue") -> int:
        """Hash the gene symbol (without value)."""
        # TODO note about hashing collisions
        return hash(self.symbol)

    def __eq__(self: "GeneValue", other: Any) -> bool:  # noqa: ANN401
        """Compare the gene symbol (without value)."""
        if isinstance(other, GeneValue):
            return self.symbol == other.symbol
        return False


class GeneDatabase(BaseModel):
    """Gene database schema."""

    name: str
    shortname: str
    species: str
    data: datetime.datetime
    precision: int


class GeneDatabaseRow(BaseModel):
    """Gene database schema for database row."""

    gdb_id: int
    gdb_name: str
    sp_id: int
    gdb_shortname: str
    gdb_date: str
    gdb_precision: int
    gdb_linkout_url: Optional[str]



"""Schemas relating to species."""
import datetime
from typing import Any, Optional

from pydantic import BaseModel, Json


class Species(BaseModel):
    """Species schema."""

    id: int  # noqa: A003
    name: str
    taxonomic_id: int


class SpeciesRow(BaseModel):
    """Species schema for database row."""

    sp_id: int
    sp_name: str
    sp_taxid: int
    sp_ref_gdb_id: Optional[int]
    sp_date: datetime.date
    sp_biomart_info: Optional[str]
    sp_source_data: Json[Any]



"""Module for defining schemas for batch endpoints."""
# ruff: noqa: N805, ANN001, ANN101
from typing import List, Optional

from geneweaver.core.parse.score import parse_score
from geneweaver.core.schema.messages import MessageResponse
from geneweaver.core.schema.score import GenesetScoreType
from geneweaver.core.schema.gene import GeneValue
from pydantic import BaseModel, validator

# Header characters which DO NOT need to be space separated.
HEADER_CHARACTERS = {
    ":": "abbreviation",
    "=": "name",
    "+": "description",
    "@": "species",
    "!": "score",
    "%": "gene_id_type",
    "~": "ontology",
}

# Header characters which DO need to be space separated. This is because these
# characters can also be seen in the values section of the batch file.
SPACE_SEPARATED_HEADER_CHARACTERS = {
    "P": "pubmed_id",
    "A": "private",
    "T": "curation_id",
    "U": "user_id",
    "D": "attribution_id",
}

# Characters which should be ignored in the batch file.
IGNORE_CHARACTERS = {
    "#": "comment",
    " ": "space",
}

# Batch characters
# TODO: Switch to this 3.9+ syntax.
# For now, we use the 3.5+ syntax.
CHAR_MAP = {
    **HEADER_CHARACTERS,
    **SPACE_SEPARATED_HEADER_CHARACTERS,
    **IGNORE_CHARACTERS,
}

# Inverse character map for converting from header name to character.
INV_CHAR_MAP = {v: k for k, v in CHAR_MAP.items()}

# Characters denoting required header information for all genesets.
REQUIRED_HEADERS = (":", "=", "+")


class BatchResponse(BaseModel):
    """Class for defining a response containing batch results."""

    genesets: List[int]
    messages: MessageResponse


GeneValueInput = GeneValue


class GenesetValue(BaseModel):
    """Class for defining a geneset value as processed."""

    ode_gene_id: str
    value: float
    ode_ref_id: str
    threshold: bool


class BatchUploadGeneset(BaseModel):
    """Class for defining a geneset uploaded using a batch upload file."""

    score: GenesetScoreType
    # TODO: Use enum from core
    species: str
    gene_id_type: str
    pubmed_id: str
    private: bool = True
    curation_id: Optional[int] = None
    abbreviation: str
    name: str
    description: str = ""
    values: List[GenesetValueInput]

    @validator("score", pre=True)
    def initialize_score(cls, v) -> GenesetScoreType:
        """Initialize score type."""
        if isinstance(v, GenesetScoreType):
            return v
        return parse_score(v)

    @validator("private", pre=True)
    def private_to_bool(cls, v) -> bool:
        """Convert private str to bool."""
        return v.lower() != "public"

    @validator("curation_id", pre=True)
    def curation_id_to_int(cls, v, values) -> int:
        """Initialize curation id based on `private` value."""
        if not v:
            # If the geneset is private, it should be set to have
            # curation tier 5, otherwise it should be set to have
            # curation tier 4.
            # It should default to private if not specified.
            return 5 if values.get("private", True) else 4
        return v



"""Enum definitions for parsing."""
from enum import Enum


class FileType(str, Enum):
    """Enum for file types."""

    TEXT = "txt"
    EXCEL = "xlsx"
    CSV = "csv"


class GeneweaverFileType(str, Enum):
    """Enum for geneweaver specific file types."""

    BATCH = "batch"
    VALUES = "values"



"""User related schemas."""
import datetime
from typing import List, Optional

from geneweaver.core.enum import AdminLevel
from geneweaver.core.schema.stubgenerator import StubGenerator
from pydantic import BaseModel


class UserRequiredFields(BaseModel):
    """User schema for required fields."""

    id: int  # noqa: A003
    email: str
    prefs: str = "{}"
    is_guest: bool = False


class User(UserRequiredFields):
    """User schema."""

    first_name: Optional[str] = None
    last_name: Optional[str] = None
    password: Optional[str] = None
    admin: AdminLevel = AdminLevel.NORMAL_USER
    last_seen: Optional[datetime.datetime] = None
    create: Optional[datetime.date] = None
    ip_address: Optional[str] = None
    api_key: Optional[str] = None
    sso_id: Optional[str] = None


class UserFull(User):
    """User schema with full information."""

    groups: List[str]
    stubgenerators: List[StubGenerator]



"""Code to check if geneset values fall within a score threshold.

This module contains code to check if geneset values fall within a score threshold.
This is required for inserting geneset values into the database.

To check if geneset values fall within a score threshold, use the `check_threshold`
function. You can pass multiple values in a list, allowing you to check an entire
geneset at once.

Or, you can pass a single value to check a single geneset, just
encapsulate the value in a list.

High-level functions:
- `check_threshold`: Check if geneset values fall within a score threshold.

Utility functions:
- `one_sided_threshold`: Check if a value falls within a one-sided threshold.
- `two_sided_threshold`: Check if a value falls within a two-sided threshold.
"""
from typing import List

from geneweaver.core.schema.batch import GenesetValueInput
from geneweaver.core.schema.score import GenesetScoreType, ScoreType


def check_threshold(
    geneset_values: List[GenesetValueInput], score: GenesetScoreType
) -> List[bool]:
    """Check to see if the geneset values fall within the score threshold.

    This function checks to see if the geneset values fall within the score threshold.

    :param geneset_values: The list of geneset values to check.
    :param score: The score type to check against.

    :returns: A list of booleans representing whether each geneset value falls within
    the score threshold.
    """
    score_type = score.score_type
    if score_type is ScoreType.BINARY:
        return [True] * len(geneset_values)

    elif score_type is ScoreType.P_VALUE or score_type is ScoreType.Q_VALUE:
        return [
            one_sided_threshold(gsv.value, score.threshold) for gsv in geneset_values
        ]

    elif score_type is ScoreType.EFFECT or score_type is ScoreType.CORRELATION:
        return [
            two_sided_threshold(gsv.value, score.threshold_low, score.threshold)
            for gsv in geneset_values
        ]


def one_sided_threshold(value: float, threshold: float) -> bool:
    """Check if a value falls within a one-sided threshold.

    This function checks if a value falls within a one-sided threshold. For example, if
    the threshold is 0.05, then this function will return True if the value is less than
    or equal to 0.05.

    :param value: The value to check.
    :param threshold: The threshold to check against.

    :returns: True if the value falls within the threshold, False otherwise.
    """
    return value <= threshold


def two_sided_threshold(value: float, threshold_low: float, threshold: float) -> bool:
    """Check if a value falls within a two-sided threshold.

    This function checks if a value falls within a two-sided threshold. For example, if
    the threshold_low is 0.1 and the threshold is 0.05, then this function will return
    True if the value is between 0.1 and 0.05.

    :param value: The value to check.
    :param threshold_low: The lower threshold to check against.
    :param threshold: The upper threshold to check against.

    :returns: True if the value falls within the threshold, False otherwise.
    """
    # If you're coming from C, you might think this wouldn't work...
    # But Python is cool like that.
    # https://docs.python.org/3/reference/expressions.html#comparisons
    # tl;dr This is equivalent to: threshold_low <= value and value <= threshold
    return threshold_low <= value <= threshold



"""Namespace for defining schemas related to messaging."""
import enum
from typing import List, Optional

from pydantic import BaseModel


class MessageType(enum.Enum):
    """Enum for defining the type of message."""

    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


class Message(BaseModel):
    """Base class for defining a message."""

    message: str
    message_type: MessageType
    detail: Optional[str] = None


class UserMessage(Message):
    """Class for defining a message for a user."""


class SystemMessage(Message):
    """Class for defining a message for the system."""


class MessageResponse(BaseModel):
    """Class for defining a response containing messages."""

    user_messages: Optional[List[UserMessage]] = None
    system_messages: Optional[List[SystemMessage]] = None



"""Publication schemas."""
from typing import Optional

from pydantic import BaseModel


class PublicationInfo(BaseModel):
    """Publication upload schema (no ID)."""

    authors: str
    title: str
    abstract: str
    journal: Optional[str] = None
    volume: Optional[str] = None
    pages: str
    month: str
    year: int
    pubmed_id: int


class Publication(PublicationInfo):
    """Publication schema (with ID)."""

    id: int  # noqa: A003



"""Functions for parsing pandas dataframes.

The top-level functions are (in-order):
- dataframe_to_gene_values
- dataframe_to_gene_values_by_name
- dataframe_to_gene_values_by_idx
"""
from typing import List, Tuple

import pandas as pd
from geneweaver.core.parse.numpy import (
    SYMBOL_KEYS,
    VALUE_KEYS,
    ndarray_to_gene_values_by_idx,
)
from geneweaver.core.schema.batch import GenesetValueInput


def dataframe_to_gene_values(geneset_df: pd.DataFrame) -> List[GenesetValueInput]:
    """Convert a pandas dataframe to a list of GeneValueInput objects.

    This function will try to map via labels first, then by index.

    :param geneset_df: The pandas dataframe to convert.
    :raises ValueError: If there's a problem mapping the dataframe.
    :return: A set of the values in the dataframe.
    """
    check_input_shape_and_type(geneset_df)
    try:
        return dataframe_to_gene_values_named(geneset_df)
    except ValueError:
        return dataframe_to_gene_values_by_idx(geneset_df)


def dataframe_to_gene_values_by_idx(
    geneset_df: pd.DataFrame,
) -> List[GenesetValueInput]:
    """Convert a pandas dataframe to a list of GeneValueInput objects.

    The pandas dataframe must be a 2-dimensional dataframe with 2 columns. The first
    column will be mapped to the GenesetValueInput.symbol attribute, and the second
    column will be mapped to the GenesetValueInput.value attribute.

    :param geneset_df: The pandas dataframe to convert.
    :return: A set of the values in the dataframe.
    """
    return ndarray_to_gene_values_by_idx(geneset_df.values)


def dataframe_to_gene_values_named(geneset_df: pd.DataFrame) -> List[GenesetValueInput]:
    """Convert a pandas dataframe to a list of GeneValueInput objects using labels.

    :param geneset_df: The pandas dataframe to convert.
    :raises ValueError: If there's a problem mapping the dataframe using labels.
    :return: A set of the values in the dataframe.
    """
    symbol_key, value_key = map_dataframe_labels_to_gene_value_attr(geneset_df)

    return [
        GenesetValueInput(symbol=row[symbol_key], value=row[value_key])
        for _, row in geneset_df.iterrows()
    ]


def check_input_shape_and_type(input_df: pd.DataFrame) -> None:
    """Check the shape and type of the input dataframe.

    :raises ValueError: If the input dataframe is not a pandas dataframe or if it is not
    at least 2-dimensional.
    """
    if not isinstance(input_df, pd.DataFrame):
        raise ValueError("Input is not a pandas dataframe")

    if input_df.ndim < 2:
        raise ValueError("Input must be at least a 2-dimensional dataframe")


def map_dataframe_labels_to_gene_value_attr(
    geneset_dataframe: pd.DataFrame,
) -> Tuple[str, str]:
    """Map DataFrame column names to GenesetValueInput attributes.

    This function checks if a DataFrame has certain column names, and if it does,
    it maps these to a GenesetValueInput object.

    In order, it will map ("Symbol", "GeneID", "Gene_ID", "Gene ID") to the
    GenesetValueInput symbol attribute.

    In order, it will map
    ("Value", "Score", "PValue", "QValue", "Effect", "Correlation") to the
    GenesetValueInput value attribute.

    If it cannot map to both symbol and value, it will raise an error.

    :param geneset_dataframe: The pandas DataFrame to map.

    :raises ValueError: If the DataFrame column names cannot be mapped to
    GenesetValueInput attributes.
    :return: A tuple, where the first value is the key mapping to the symbol attribute
    and the second value is the key mapping to the value attribute.
    """
    symbol_key, value_key = None, None
    for key in SYMBOL_KEYS:
        if key in geneset_dataframe.columns:
            symbol_key = key
            break

    for key in VALUE_KEYS:
        if key in geneset_dataframe.columns:
            value_key = key
            break

    if symbol_key is None or value_key is None:
        raise ValueError(
            "Could not map DataFrame column names to GenesetValueInput attributes"
        )

    return symbol_key, value_key



"""Parse Excel files for use by the client library."""
from typing import Dict, List, Optional, Tuple, Union
from zipfile import BadZipFile

from geneweaver.core.types import StringOrPath
from openpyxl import load_workbook
from openpyxl.utils.exceptions import InvalidFileException
from openpyxl.worksheet.worksheet import Worksheet


def get_sheet_names(file_path: StringOrPath) -> List[str]:
    """Retrieve the names of all sheets in an Excel file.

    :param file_path: The path to the Excel file.

    :returns: A list of strings representing the names of all sheets in the file.
    """
    try:
        workbook = load_workbook(filename=file_path, read_only=True)
    except (InvalidFileException, BadZipFile) as e:
        raise ValueError(str(e)) from e
    return workbook.sheetnames


def get_sheet(file_path: StringOrPath, sheet_name: str = None) -> Worksheet:
    """Load a workbook and return a sheet.

    This function opens an Excel workbook and returns the specified sheet if a sheet
    name is provided.

    If no sheet name is provided, it returns the active sheet.

    :param file_path: Path to the Excel (.xlsx) file.
    :param sheet_name: The name of the sheet to open. If not provided, the active sheet
    is returned.

    :returns: The requested worksheet from the Excel workbook.
    """
    workbook = load_workbook(filename=file_path, read_only=True)
    if sheet_name:
        return workbook[sheet_name]
    else:
        return workbook.active


def find_header(
    file_path: StringOrPath, max_rows_to_check: int = 5, sheet_name: str = None
) -> Tuple[bool, int]:
    """Determine if a given Excel worksheet has a header row.

    This function will check up to 'max_rows_to_check' rows from the top to find a
    header row. It assumes that a row can be interpreted as a header if none of its
    values are numeric, if it has more than one column, and if the following row has the
    same number of columns that it does.

    :param file_path: Path to the Excel (.xlsx) file.
    :param max_rows_to_check: The number of rows to check from the top to find a header
    row. Default is 5.
    :param sheet_name: The name of the sheet to open. If not provided, the active sheet
    is returned.

    :returns:
        First index - True if a header row is found, False otherwise.
        Second index - The index of the header row if found, otherwise -1.
    """
    sheet = get_sheet(file_path, sheet_name)
    for i, row in enumerate(sheet.iter_rows(max_row=max_rows_to_check), start=1):
        values = [cell.value for cell in row]
        if (
            len(values) > 1
            and all(value is not None for value in values)
            and all(not isinstance(value, (int, float)) for value in values)
        ):
            next_values = [cell.value for cell in sheet[i + 1]]
            if next_values and len(values) == len(next_values):
                return True, i - 1
    return False, -1


def has_header(
    file_path: str, max_rows_to_check: int = 5, sheet_name: str = None
) -> bool:
    """Summary function to return true/false for if a header is found by `find_header`.

    :param file_path: Path to the Excel (.xlsx) file.
    :param max_rows_to_check: The number of rows to check from the top to find a
    header row.

    :returns:  True if a header row is found, False otherwise.
    """
    return find_header(file_path, max_rows_to_check, sheet_name)[0]


def get_headers(
    file_path: StringOrPath, sheet_name: Optional[str] = None
) -> Tuple[List[str], int]:
    """Read the headers of an Excel (.xlsx) file.

    Read the first row from an Excel (.xlsx) file and return it as a list of strings.

    Assumes that the first row of the Excel file is the header row.

    :param file_path: Path to the Excel file.
    :param sheet_name: Name of the sheet to read from. If not provided, the function
    will read from the active sheet.

    :returns: List of column names from the CSV file, and the index of the
    header row.
    """
    headers = []
    file_has_header, header_idx = find_header(file_path, sheet_name=sheet_name)
    if file_has_header:
        headers = read_row(file_path, header_idx, sheet_name=sheet_name)
    return headers, header_idx


def read_row(file_path: str, row_idx: int = 0, sheet_name: str = None) -> List[str]:
    """Get the contents of a row from an Exel (.xlsx) file.

    :param file_path: The file path to the CSV file.
    :param row_idx: The index of the row from which to read the headers. Defaults to 0.
    :param sheet_name: Name of the sheet to read from. If not provided, the function
    will read from the active sheet.

    :returns: The contents of a row

    :raises ValueError: If the file is empty or does not contain enough rows.
    """
    sheet = get_sheet(file_path, sheet_name)
    row_idx += 1
    try:
        return [cell.value for cell in sheet[row_idx]]
    except IndexError as e:
        raise ValueError(f"File does not contain a row at index {row_idx}") from e


def read_to_dict(
    file_path: StringOrPath, start_row: int = 0, sheet_name: Optional[str] = None
) -> List[Dict[str, Union[str, int]]]:
    """Parse an Excel file into a list of dictionaries.

    Parse an Excel file into a list of dictionaries, with keys as column names and
    values as column values.

    :param file_path: The file path to the Excel file.
    :param start_row: The row number to start reading from (0-indexed).
                               This row will be used as the header.
                               Defaults to 0.
    :param sheet_name: Name of the sheet to read from. If not provided, the function
    will read from the active sheet.

    :returns: A list of dictionaries, where each dictionary
    represents a row from the Excel file.
    """
    headers = read_row(file_path, start_row, sheet_name=sheet_name)
    sheet = get_sheet(file_path, sheet_name)
    data = []
    for row in sheet.iter_rows(min_row=start_row + 2):
        row_data = {h: cell.value for h, cell in zip(headers, row)}  # noqa: B905
        data.append(row_data)
    return data


def read_to_dict_n_rows(
    file_path: str, n: int, start_row: int = 0, sheet_name: str = None
) -> List[Dict[str, str]]:
    """Parse n lines of an Excel file into a list of dictionaries.

    Parse an Excel file into a list of dictionaries, with keys as column names and
    values as column values. This function will only return the first n rows of data.

    :param file_path: The file path to the Excel file.
    :param n: The number of rows of data to return.
    :param start_row: The row number to start reading from (0-indexed).
                           This row will be used as the header.
                           Defaults to 0.
    :param sheet_name: Name of the sheet to read from. If not provided, the function
    will read from the active sheet.

    :returns: A list of dictionaries, where each dictionary represents
    a row from the Excel file.
    """
    headers = read_row(file_path, start_row, sheet_name=sheet_name)
    sheet = get_sheet(file_path, sheet_name)
    data = []
    for row in sheet.iter_rows(min_row=start_row + 2, max_row=start_row + 1 + n):
        row_data = {h: cell.value for h, cell in zip(headers, row)}  # noqa: B905
        data.append(row_data)
    return data



"""Gene fixtures for testing."""
from geneweaver.core.schema.gene import Gene, GeneDatabase, GeneDatabaseRow, GeneRow

GENE_ROWS = [
    GeneRow(
        ode_gene_id=606485,
        ode_ref_id="100683049",
        gdb_id=1,
        sp_id=11,
        ode_pref=False,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=606486,
        ode_ref_id="100683125",
        gdb_id=1,
        sp_id=11,
        ode_pref=False,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=606487,
        ode_ref_id="100683204",
        gdb_id=1,
        sp_id=11,
        ode_pref=False,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=606488,
        ode_ref_id="100683346",
        gdb_id=1,
        sp_id=11,
        ode_pref=False,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=606489,
        ode_ref_id="100856548",
        gdb_id=1,
        sp_id=11,
        ode_pref=False,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=606490,
        ode_ref_id="111094793",
        gdb_id=1,
        sp_id=11,
        ode_pref=False,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=606491,
        ode_ref_id="111094899",
        gdb_id=1,
        sp_id=11,
        ode_pref=False,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=606492,
        ode_ref_id="111094906",
        gdb_id=1,
        sp_id=11,
        ode_pref=False,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=606493,
        ode_ref_id="111094786",
        gdb_id=1,
        sp_id=11,
        ode_pref=False,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=606494,
        ode_ref_id="612605",
        gdb_id=1,
        sp_id=11,
        ode_pref=False,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=92095,
        ode_ref_id="A1BG",
        gdb_id=7,
        sp_id=2,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=92085,
        ode_ref_id="A1BG-AS1",
        gdb_id=7,
        sp_id=2,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=75095,
        ode_ref_id="A1CF",
        gdb_id=7,
        sp_id=2,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=74678,
        ode_ref_id="A2M",
        gdb_id=7,
        sp_id=2,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=157301,
        ode_ref_id="7SK",
        gdb_id=7,
        sp_id=3,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=154481,
        ode_ref_id="7SK",
        gdb_id=7,
        sp_id=3,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=153231,
        ode_ref_id="7SK",
        gdb_id=7,
        sp_id=3,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=151248,
        ode_ref_id="7SK",
        gdb_id=7,
        sp_id=3,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=191467,
        ode_ref_id="a008504",
        gdb_id=7,
        sp_id=4,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=193962,
        ode_ref_id="a0264234",
        gbd_id=7,
        sp_id=4,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=190256,
        ode_ref_id="a0264239",
        gbd_id=7,
        sp_id=4,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=194073,
        ode_ref_id="a0274750",
        gbd_id=7,
        sp_id=4,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=290564,
        ode_ref_id="7SLRNA:CR32864",
        gbd_id=7,
        sp_id=5,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=290563,
        ode_ref_id="7SLRNA:CR42652",
        gbd_id=7,
        sp_id=5,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=461964,
        ode_ref_id="825-Oak",
        gbd_id=7,
        sp_id=5,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
    GeneRow(
        ode_gene_id=290329,
        ode_ref_id="82Y",
        gbd_id=7,
        sp_id=5,
        ode_pref=True,
        ode_date="2020-05-05",
        old_ode_gene_ids=None,
    ),
]

GENES = [
    Gene(
        id=606485,
        reference_id="100683049",
        gene_database="Entrez",
        species="Mus musculus",
        preferred=False,
        date="2020-05-05",
    ),
    Gene(
        id=606486,
        reference_id="100683125",
        gene_database="Entrez",
        species="Mus musculus",
        preferred=False,
        date="2020-05-05",
    ),
    Gene(
        id=606487,
        reference_id="100683204",
        gene_database="Entrez",
        species="Mus musculus",
        preferred=False,
        date="2020-05-05",
    ),
    Gene(
        id=606488,
        reference_id="100683346",
        gene_database="Entrez",
        species="Mus musculus",
        preferred=False,
        date="2020-05-05",
    ),
    Gene(
        id=606489,
        reference_id="100856548",
        gene_database="Entrez",
        species="Mus musculus",
        preferred=False,
        date="2020-05-05",
    ),
    Gene(
        id=606490,
        reference_id="111094793",
        gene_database="Entrez",
        species="Mus musculus",
        preferred=False,
        date="2020-05-05",
    ),
    Gene(
        id=606491,
        reference_id="111094899",
        gene_database="Entrez",
        species="Mus musculus",
        preferred=False,
        date="2020-05-05",
    ),
    Gene(
        id=606492,
        reference_id="111094906",
        gene_database="Entrez",
        species="Mus musculus",
        preferred=False,
        date="2020-05-05",
    ),
    Gene(
        id=606493,
        reference_id="111094786",
        gene_database="Entrez",
        species="Mus musculus",
        preferred=False,
        date="2020-05-05",
    ),
    Gene(
        id=606494,
        reference_id="612605",
        gene_database="Entrez",
        species="Mus musculus",
        preferred=False,
        date="2020-05-05",
    ),
    Gene(
        id=92095,
        reference_id="A1BG",
        gene_database="Gene Symbol",
        species="Homo sapiens",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=92085,
        reference_id="A1BG-AS1",
        gene_database="Gene Symbol",
        species="Homo sapiens",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=75095,
        reference_id="A1CF",
        gene_database="Gene Symbol",
        species="Homo sapiens",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=74678,
        reference_id="A2M",
        gene_database="Gene Symbol",
        species="Homo sapiens",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=157301,
        reference_id="7SK",
        gene_database="Gene Symbol",
        species="Rattus norvegicus",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=154481,
        reference_id="7SK",
        gene_database="Gene Symbol",
        species="Rattus norvegicus",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=153231,
        reference_id="7SK",
        gene_database="Gene Symbol",
        species="Rattus norvegicus",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=151248,
        reference_id="7SK",
        gene_database="Gene Symbol",
        species="Rattus norvegicus",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=191467,
        reference_id="a008504",
        gene_database="Gene Symbol",
        species="Danio rerio",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=193962,
        reference_id="a0264234",
        gene_database="Gene Symbol",
        species="Danio rerio",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=190256,
        reference_id="a0264239",
        gene_database="Gene Symbol",
        species="Danio rerio",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=194073,
        reference_id="a0274750",
        gene_database="Gene Symbol",
        species="Danio rerio",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=290564,
        reference_id="7SLRNA:CR32864",
        gene_database="Gene Symbol",
        species="Drosophila melanogaster",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=290563,
        reference_id="7SLRNA:CR42652",
        gene_database="Gene Symbol",
        species="Drosophila melanogaster",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=461964,
        reference_id="825-Oak",
        gene_database="Gene Symbol",
        species="Drosophila melanogaster",
        preferred=True,
        date="2020-05-05",
    ),
    Gene(
        id=290329,
        reference_id="82Y",
        gene_database="Gene Symbol",
        species="Drosophila melanogaster",
        preferred=True,
        date="2020-05-05",
    ),
]

GENE_DATABASE_ROWS = [
    GeneDatabaseRow(
        gdb_id=8,
        gdb_name="Unannotated",
        sp_id=0,
        gdb_shortname="unannotated",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=4,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=4,
        gdb_name="Ensembl Transcript",
        sp_id=0,
        gdb_shortname="ensembl transcript",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=2,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=3,
        gdb_name="Ensembl Protein",
        sp_id=0,
        gdb_shortname="ensembl protein",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=2,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=7,
        gdb_name="Gene Symbol",
        sp_id=0,
        gdb_shortname="symbol",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=4,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=10,
        gdb_name="MGI",
        sp_id=1,
        gdb_shortname="mgi",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=1,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=11,
        gdb_name="HGNC",
        sp_id=2,
        gdb_shortname="hgnc",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=1,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=12,
        gdb_name="RGD",
        sp_id=3,
        gdb_shortname="rgd",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=1,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=13,
        gdb_name="ZFIN",
        sp_id=4,
        gdb_shortname="zfin",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=1,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=14,
        gdb_name="FlyBase",
        sp_id=5,
        gdb_shortname="flybase",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=1,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=1,
        gdb_name="Entrez",
        sp_id=0,
        gdb_shortname="entrez",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=2,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=2,
        gdb_name="Ensembl Gene",
        sp_id=0,
        gdb_shortname="ensembl",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=2,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=5,
        gdb_name="Unigene",
        sp_id=0,
        gdb_shortname="unigene",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=3,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=15,
        gdb_name="Wormbase",
        sp_id=8,
        gdb_shortname="wormbase",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=1,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=16,
        gdb_name="SGD",
        sp_id=9,
        gdb_shortname="sgd",
        gdb_date="2011-03-14 14:17:13.49449",
        gdb_precision=1,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=17,
        gdb_name="miRBase",
        sp_id=0,
        gdb_shortname="mirbase",
        gdb_date="2011-09-12 14:38:49.943681",
        gdb_precision=2,
        gdb_linkout_url=None,
    ),
    GeneDatabaseRow(
        gdb_id=20,
        gdb_name="CGNC",
        sp_id=10,
        gdb_shortname="cgnc",
        gdb_date="2015-06-05 18:05:13.012876",
        gdb_precision=1,
        gdb_linkout_url=None,
    ),
]

GENE_DATABASES = [
    GeneDatabase(
        name="Unannotated",
        shortname="unannotated",
        species="all",
        data="2011-03-14 14:17:13.49449",
        precision=4,
    ),
    GeneDatabase(
        name="Ensembl Transcript",
        shortname="ensembl transcript",
        species="all",
        data="2011-03-14 14:17:13.49449",
        precision=2,
    ),
    GeneDatabase(
        name="Ensembl Protein",
        shortname="ensembl protein",
        species="all",
        data="2011-03-14 14:17:13.49449",
        precision=2,
    ),
    GeneDatabase(
        name="Gene Symbol",
        shortname="symbol",
        species="all",
        data="2011-03-14 14:17:13.49449",
        precision=4,
    ),
    GeneDatabase(
        name="MGI",
        shortname="mgi",
        species="Mus musculus",
        data="2011-03-14 14:17:13.49449",
        precision=1,
    ),
    GeneDatabase(
        name="HGNC",
        shortname="hgnc",
        species="Homo sapiens",
        data="2011-03-14 14:17:13.49449",
        precision=1,
    ),
    GeneDatabase(
        name="RGD",
        shortname="rgd",
        species="Rattus norvegicus",
        data="2011-03-14 14:17:13.49449",
        precision=1,
    ),
    GeneDatabase(
        name="ZFIN",
        shortname="zfin",
        species="Danio rerio",
        data="2011-03-14 14:17:13.49449",
        precision=1,
    ),
    GeneDatabase(
        name="FlyBase",
        shortname="flybase",
        species="Drosophila melanogaster",
        data="2011-03-14 14:17:13.49449",
        precision=1,
    ),
    GeneDatabase(
        name="Entrez",
        shortname="entrez",
        species="all",
        data="2011-03-14 14:17:13.49449",
        precision=2,
    ),
    GeneDatabase(
        name="Ensembl Gene",
        shortname="ensembl",
        species="all",
        data="2011-03-14 14:17:13.49449",
        precision=2,
    ),
    GeneDatabase(
        name="Unigene",
        shortname="unigene",
        species="all",
        data="2011-03-14 14:17:13.49449",
        precision=3,
    ),
    GeneDatabase(
        name="Wormbase",
        shortname="wormbase",
        species="Caenorhabditis elegans",
        data="2011-03-14 14:17:13.49449",
        precision=1,
    ),
    GeneDatabase(
        name="SGD",
        shortname="sgd",
        species="Saccharomyces cerevisiae",
        data="2011-03-14 14:17:13.49449",
        precision=1,
    ),
    GeneDatabase(
        name="miRBase",
        shortname="mirbase",
        species="all",
        data="2011-09-12 14:38:49.943681",
        precision=2,
    ),
    GeneDatabase(
        name="CGNC",
        shortname="cgnc",
        species="Gallus gallus",
        data="2015-06-05 18:05:13.012876",
        precision=1,
    ),
]



"""Schema module for GeneWeaver."""
# ruff: noqa: F401
from .gene import Gene, GeneValue
from .geneset import Geneset, GenesetGenes, GenesetUpload
from .group import Group, UserAdminGroup
from .project import Project, ProjectCreate
from .publication import Publication, PublicationInfo
from .stubgenerator import StubGenerator
from .user import User



"""Parse batch geneset files.

This module provides a set of functions for processing batch geneset files and
their contents. It includes functionality to check the type of a file,
process lines of content, read headers and values, update and reset headers,
and create Geneset objects for batch upload.

File Type Identification Function:
- is_batch_file: Check if the provided file contents are a valid batch file.

Line Processing Function:
- process_lines: Process each line of content to build and return a list of
BatchUploadGeneset instances.

Header Processing Functions:
- read_header: Process a header line and update the corresponding state variables
accordingly.
- update_header: Update the header dictionary based on a key-value pair from the
processed line.
- reset_required_header_values: Reset required values in a header dictionary.
- check_has_required_header_values: Check if a header dictionary has required values.
- process_header_line: Process a header line into a key-value pair.
- read_single_prefix_header: Read a single-prefix header line into a key-value pair.
- read_space_separated_header: Read a space-separated header line into a key-value pair.

Value Processing Functions:
- read_values: Read a line assuming it's a value, and updates the reading mode if
necessary.
- process_value_line: Process a value line into a key-value pair.

Geneset Construction Function:
- create_geneset: Create a Geneset object from a batch upload.

Geneset Finalization Function:
- finalize_processed_geneset: Add the current geneset to the list and prepares for
processing the next one.

Exception Checking Function:
- string_has_newlines: Check if a string has newline characters.

Exceptions:
- UnsupportedFileTypeError: Raised if the file is neither a batch file nor a geneset
values file.
- MultiLineStringError: Raised if a value row is encountered with return characters.
- InvalidBatchValueLineError: Raised if a value row is encountered with more than two
values.
- NotAHeaderRowError: Raised when a header line is expected but not found.
- MissingRequiredHeaderError: Raised when any of the required keys are missing in the
header.
- IgnoreLineError: Raised if the line is to be ignored based on its prefix.
"""

from enum import Enum
from typing import List, Tuple

from geneweaver.core.parse.enum import GeneweaverFileType
from geneweaver.core.parse.exceptions import (
    IgnoreLineError,
    InvalidBatchValueLineError,
    MissingRequiredHeaderError,
    MultiLineStringError,
    NotAHeaderRowError,
    UnsupportedFileTypeError,
)
from geneweaver.core.schema.batch import BatchUploadGeneset, GenesetValueInput

# Header characters which DO NOT need to be space separated.
HEADER_CHARACTERS = {
    ":": "abbreviation",
    "=": "name",
    "+": "description",
    "@": "species",
    "!": "score",
    "%": "gene_id_type",
    "~": "ontology",
}

# Header characters which DO need to be space separated. This is because these
# characters can also be seen in the values section of the batch file.
SPACE_SEPARATED_HEADER_CHARACTERS = {
    "P": "pubmed_id",
    "A": "private",
    "T": "curation_id",
    "U": "user_id",
    "D": "attribution_id",
}

# Characters which should be ignored in the batch file.
IGNORE_CHARACTERS = {
    "#": "comment",
    " ": "space",
}

# Characters denoting required header information for all genesets.
REQUIRED_HEADERS = (":", "=", "+")


class ReadMode(Enum):
    """Enum to keep track of what part of the batch file is being read."""

    HEADER = "header"
    CONTENT = "content"


def is_batch_file(contents: str) -> GeneweaverFileType:
    """Check if the provided file contents are a valid batch file.

    We will consider a file to be a batch file if we encounter a header row before
    encountering a value row. If we encounter a value row before a header row, we will
    consider the file to be a geneset file.

    Processing stops as soon as a header or value row is encountered. If each line
    contains an ignore character (e.g. a comment), the entire file will be read.

    :param contents: The contents of the file to be checked.

    :raises: UnsupportedFileTypeError: If the file is neither a batch file nor a
    geneset values file. Or if any of the following errors are encountered during
    reading:
        - MultiLineStringError: If a value row is encountered with return characters.
        - InvalidBatchValueLineError: If a value row is encountered with more than two
        values (the gene identifier and the score).

    :returns: GeneweaverFileType.BATCH if the file is a batch file,
    GeneweaverFileType.VALUES otherwise.
    """
    for line in contents.splitlines():
        try:
            _ = process_header_line(line)
            return GeneweaverFileType.BATCH

        except NotAHeaderRowError:
            try:
                _ = process_value_line(line)
                return GeneweaverFileType.VALUES

            except InvalidBatchValueLineError as e:
                raise UnsupportedFileTypeError() from e

            except MultiLineStringError as e:
                raise UnsupportedFileTypeError() from e

        except IgnoreLineError:
            continue

    raise UnsupportedFileTypeError()


def process_lines(contents: str) -> List[BatchUploadGeneset]:
    """Process each line of content to build and return a list of BatchUploadGeneset.

    The function iterates over each line in the provided content. Depending on the
    current read mode and the content of the line, the function performs one of the
    following actions:

    - If the line contains header information and the current read mode is 'HEADER',
      it updates the header information for the current geneset.
    - If the line contains header information and the current read mode is 'CONTENT',
      it finalizes the current geneset and prepares for a new one. Header information
      encountered after the presence of geneset value information indicates that the
      current geneset is complete and should be appended to the genesets list.
    - If the line does not contain header information it reads the line as geneset value
      information, appends the value to the current geneset values and
      - will switch the read mode to 'CONTENT' if it is currently 'HEADER'.
    - If the line contains an ignored character, it ignores the line and continues.

    After processing all lines, the function appends the final geneset (created from the
    last header and value lines) to the genesets list.

    :param contents: The contents of the batch file to be processed.

    :returns: A list of genesets created from the processed batch file.
    """
    genesets, header, current_geneset_values, read_mode = [], {}, [], ReadMode.HEADER

    for line in contents.splitlines():
        try:
            genesets, current_geneset_values, header, read_mode = read_header(
                line, header, current_geneset_values, read_mode, genesets
            )

        except NotAHeaderRowError:
            current_geneset_values, read_mode = read_values(
                line, header, current_geneset_values, read_mode
            )

        except IgnoreLineError:
            continue

    genesets.append(create_geneset(header, current_geneset_values))

    return genesets


def read_header(
    line: str,
    header: dict,
    current_geneset_values: list,
    read_mode: ReadMode,
    genesets: List[BatchUploadGeneset],
) -> Tuple[list[BatchUploadGeneset], list, dict, ReadMode]:
    """Process a header line and update the corresponding state variables accordingly.

    This function takes a line from a batch file, processes it as a header line and
    updates the header dictionary, the current_geneset_values list, and the read_mode.
    If the read_mode was set to CONTENT prior to reading the line, it implies that the
    previous line was a geneset value line, thus it finalizes the current geneset and
    adds it to the genesets list, and resets the header and current_geneset_values for
    the next geneset.

    :param line: The line from the batch file to be processed.
    :param header: The dict that holds the header information for the current geneset.
    :param current_geneset_values: The list of current geneset values.
    :param read_mode: The current read mode - HEADER or CONTENT.
    :param genesets: The list of already processed genesets.

    :returns: A tuple containing:
         0. the updated list of processed genesets,
         1. the current geneset values,
         2. the updated header, and the updated read mode.

    :raises NotAHeaderRowError:Raised when a header line is expected but not found.
    MissingRequiredHeaderError
        If any of the required keys are missing in the header before trying to finalize
        the geneset (when switching from content to header).
    NotAHeaderRowError
        If the line does not follow the expected header format.
    """
    key, value = process_header_line(line)

    if read_mode is ReadMode.CONTENT:
        genesets, current_geneset_values, header = finalize_processed_geneset(
            genesets, header, current_geneset_values
        )
        read_mode = ReadMode.HEADER

    header = update_header(key, value, header)
    return genesets, current_geneset_values, header, read_mode


def finalize_processed_geneset(
    genesets: List[BatchUploadGeneset], header: dict, current_geneset_values: list
) -> Tuple[List[BatchUploadGeneset], list, dict]:
    """Add the current geneset to the list and prepares for processing the next one.

    This function first checks that the header contains all required values. Then, it
    creates a new geneset using the header and the current_geneset_values list and
    appends it to the genesets list. It also resets the current_geneset_values list and
    the header for the next geneset.

    :param genesets: A list of BatchUploadGeneset instances that have been processed.
    :param header: A dictionary representing the current header data.
    :param current_geneset_values: A list of GenesetValueInput instances representing
    the current geneset values.

    :returns: A tuple containing three elements:
        0. The updated list of BatchUploadGeneset instances.
        1. The reset list for the next geneset values.
        2. The reset header dictionary for the next geneset.

    :raises MissingRequiredHeaderError:If any of the required keys are missing in the
    header before trying to finalize the geneset.
    """
    check_has_required_header_values(header)
    genesets.append(create_geneset(header, current_geneset_values))
    return genesets, [], reset_required_header_values(header)


def update_header(key: str, value: str, header: dict) -> dict:
    """Update the header dictionary based on a key-value pair from the processed line.

    The function updates the header dictionary in-place. This function is used to take
    any header-key specific action needed based on the header key type.

    If the key equals to 'description', it appends the value to the existing description
    value, if any, separated by a space. Otherwise, it simply updates or sets the value
    for the given key in the header.

    :param key: The key to update in the header, as a string.
    :param value: The new value for the given key, as a string.
    :param header: The header dictionary to update.

    :returns: The updated header dictionary.
    """
    if key == "description":
        header[key] = " ".join((header.get(key, ""), value))
    else:
        header[key] = value
    return header


def read_values(
    line: str, header: dict, current_geneset_values: list, read_mode: ReadMode
) -> Tuple[list, ReadMode]:
    r"""Read a line assuming it's a value, and updates the reading mode if necessary.

    This function assumes the given line is not a header row. It first uses the
    process_value_line function to parse the line into a symbol-value pair. Then, if the
    current read_mode is HEADER, it checks the header for required values and switches
    the read_mode to CONTENT. Finally, it appends a new GenesetValueInput
    instance to the current_geneset_values list.

    :param line: A string representing a line in the 'value' format, e.g.,
    'symbol\tvalue' or 'symbol value'. Will also accept multiple whitespace characters
    between the symbol and value.
    :param header: A dictionary representing the current header data.
    :param current_geneset_values: A list of GenesetValueInput instances representing
    the current geneset values.
    :param read_mode: A ReadMode enum value representing the current reading mode.

    :returns: A tuple where the first element is the updated list of GenesetValueInput
    instances, and the second element is the updated reading mode.

    :raises MissingRequiredHeaderError:If any of the required keys are missing in the
    header when switching from HEADER to CONTENT mode (the first time a content line is
    read after a header).
    """
    symbol, value = process_value_line(line)

    if read_mode is ReadMode.HEADER:
        check_has_required_header_values(header)
        read_mode = ReadMode.CONTENT

    current_geneset_values.append(GenesetValueInput(symbol=symbol, value=value))

    return current_geneset_values, read_mode


def create_geneset(
    header: dict, content: List[GenesetValueInput]
) -> BatchUploadGeneset:
    """Create a Geneset object for batch upload.

    This function takes a header and content, and constructs a BatchUploadGeneset object
    using these inputs.

    :param header: Dictionary containing header information.
    :param content: List containing content information.

    :returns: An object representing the Geneset for batch upload.
    """
    return BatchUploadGeneset(values=content, **header)


def reset_required_header_values(header: dict) -> dict:
    """Reset required values in a header dictionary.

    This function attempts to delete specific keys from the header dictionary.

    :param header: The header dictionary.

    :returns: The updated header dictionary.
    """
    for key in REQUIRED_HEADERS:
        try:
            del header[HEADER_CHARACTERS[key]]
        except KeyError:
            continue
    return header


def check_has_required_header_values(header: dict) -> None:
    """Check if a header dictionary has required values.

    This function checks if specific keys are present in the header dictionary.

    :param header: The header dictionary.

    :returns: None

    :raises MissingRequiredHeaderError: If any of the required keys are missing in the
    header.
    """
    if not all((HEADER_CHARACTERS[key] in header for key in REQUIRED_HEADERS)):
        raise MissingRequiredHeaderError()


def process_header_line(line: str) -> Tuple[str, str]:
    """Process a header line into a key-value pair.

    This function reads a header line, and processes it into a key-value pair based on
    its prefix.

    :param line: The line to process.

    :returns: The key-value pair.

    :raises IgnoreLineError: If the line is to be ignored based on its prefix.
    :raises NotAHeaderRowError: If the line does not follow the expected header format.
    """
    line = line.strip()

    try:
        prefix = line[0]
    except IndexError:
        raise IgnoreLineError() from None

    if prefix in IGNORE_CHARACTERS:
        raise IgnoreLineError() from None

    try:
        key, value = read_single_prefix_header(prefix, line)
    except NotAHeaderRowError:
        key, value = read_space_separated_header(prefix, line)

    return key, value


def read_single_prefix_header(prefix: str, line: str) -> Tuple[str, str]:
    """Read a single-prefix header line into a key-value pair.

    This function reads a single-prefix header line, and returns it as a key-value pair.

    :param prefix: The prefix of the line (the first character).
    :param line: The line to process.

    :returns: The key-value pair.

    :raises NotAHeaderRowError: If the line does not follow the expected single-prefix
    header format.
    """
    try:
        key = HEADER_CHARACTERS[prefix]
        value = line[1:].strip()

    except (KeyError, IndexError):
        raise NotAHeaderRowError() from None

    return key, value


def read_space_separated_header(prefix: str, line: str) -> Tuple[str, str]:
    """Read a space-separated header line into a key-value pair.

    This function reads a space-separated header line, and returns it as a key-value
    pair.

    :param prefix: The prefix of the line (the first character).
    :param line: The line to process.

    :returns: first value is the header key, second is the header value.

    :raises NotAHeaderRowError: If the line does not follow the expected space-separated
    header format.
    """
    try:
        if line[1] not in (" ", "\t"):
            raise NotAHeaderRowError()

        key = SPACE_SEPARATED_HEADER_CHARACTERS[prefix]
        value = line[2:].strip()

    except (KeyError, IndexError):
        raise NotAHeaderRowError() from None

    return key, value


def process_value_line(line: str) -> Tuple[str, str]:
    """Process a value line into a key-value pair.

    This function splits a line into two parts, and returns them as a symbol-value pair.

    :param line: The line to process.

    :returns: The first value is the gene symbol, the second is the value.

    :raises MultiLineStringError: If the line contains newline characters.
    :raises InvalidBatchValueLine: If the line does not split into exactly two parts.
    """
    if string_has_newlines(line):
        raise MultiLineStringError()

    split_line = line.split()

    if len(split_line) != 2:
        raise InvalidBatchValueLineError()

    return split_line[0], split_line[1]


def string_has_newlines(input_str: str) -> bool:
    r"""Check if a string has newline characters.

    This function checks if a string contains either "\n" or "\r" characters.

    :param input_str: The string to check.

    :returns: True if the string contains newline characters, False otherwise.
    """
    # Here we return the results of the expression evaluation. This means that it will
    # return True if either "\n" or "\r" is in the string, and False otherwise.
    return "\n" in input_str or "\r" in input_str



"""Models needed to work with the legacy API."""
from typing import List, Optional

from geneweaver.core.enum import GenesetAccess, GenesetScoreType
from pydantic import BaseModel, HttpUrl


class AddGenesetByUserPublication(BaseModel):
    """Publication schema for adding genesets by user."""

    pub_abstract: Optional[str]
    pub_authors: Optional[str]
    pub_journal: Optional[str]
    pub_pages: Optional[str]
    pub_pubmed: Optional[str]
    pub_title: Optional[str]
    pub_volume: Optional[str]
    pub_year: Optional[str]


class AddGenesetByUserBase(BaseModel):
    """Base schema for adding genesets by user."""

    gene_identifier: str
    gs_abbreviation: str
    gs_description: str
    gs_name: str
    gs_threshold_type: GenesetScoreType
    permissions: GenesetAccess
    publication: Optional[AddGenesetByUserPublication]
    select_groups: List[str]
    sp_id: str

    class Config:
        """Pydantic config."""

        use_enum_values = True


class AddGenesetByUser(AddGenesetByUserBase):
    """Schema for adding genesets by user."""

    file_text: str


class AddGenesetByUserFile(AddGenesetByUserBase):
    """Schema for adding genesets by user from file."""

    file_url: HttpUrl



from typing import List

from geneweaver.core.schema.batch import (
    HEADER_CHARACTERS,
    IGNORE_CHARACTERS,
    INV_CHAR_MAP,
    SPACE_SEPARATED_HEADER_CHARACTERS,
    BatchUploadGeneset,
)

CHAR_MAP = HEADER_CHARACTERS | SPACE_SEPARATED_HEADER_CHARACTERS | IGNORE_CHARACTERS


def format_batch_file(genesets: List[BatchUploadGeneset]) -> str:
    """Format a batch upload file from a list of genesets."""


def format_geneset(geneset: BatchUploadGeneset) -> str:
    geneset_dict = geneset.dict(exclude={"values", "curation_id"})
    "\n".join(
        (
            f"{INV_CHAR_MAP[key] if key in INV_CHAR_MAP else key}={value}"
            for key, value in geneset_dict.items()
        )
    )



"""Species fixtures for testing."""
from geneweaver.core.schema.species import Species, SpeciesRow

SPECIES = [
    Species(id=0, name="all", taxonomic_id=0),
    Species(id=1, name="Mus musculus", taxonomic_id=10090),
    Species(id=2, name="Homo sapiens", taxonomic_id=9606),
    Species(id=3, name="Rattus norvegicus", taxonomic_id=10116),
    Species(id=4, name="Danio rerio", taxonomic_id=7955),
    Species(id=5, name="Drosophila melanogaster", taxonomic_id=7227),
    Species(id=6, name="Macaca mulatta", taxonomic_id=9544),
    Species(id=8, name="Caenorhabditis elegans", taxonomic_id=6239),
    Species(id=9, name="Saccharomyces cerevisiae", taxonomic_id=559292),
    Species(id=10, name="Gallus gallus", taxonomic_id=9031),
    Species(id=11, name="Canis familiaris", taxonomic_id=9615),
]


SPECIES_ROWS = [
    SpeciesRow(sp_id=0, sp_name="", sp_tax_id=0, sp_date="2011-03-14"),
    SpeciesRow(
        sp_id=1,
        sp_name="Mus musculus",
        sp_tax_id=10090,
        sp_ref_gdb_id=10,
        sp_date="2011-03-14",
        sp_biomart_info="mmusculus_gene_ensembl  mgi_id",
        sp_source_data={},
    ),
    SpeciesRow(
        sp_id=2,
        sp_name="Homo sapiens",
        sp_tax_id=9606,
        sp_ref_gdb_id=11,
        sp_date="2011-03-14",
        sp_biomart_info="hsapiens_gene_ensembl   hgnc_id",
        sp_source_data={},
    ),
    SpeciesRow(
        sp_id=3,
        sp_name="Rattus norvegicus",
        sp_tax_id=10116,
        sp_ref_gdb_id=12,
        sp_date="2011-03-14",
        sp_biomart_info="rnorvegicus_gene_ensembl        rgd_id",
        sp_source_data={},
    ),
    SpeciesRow(
        sp_id=4,
        sp_name="Danio rerio",
        sp_tax_id=7955,
        sp_ref_gdb_id=13,
        sp_date="2011-03-14",
        sp_biomart_info="drerio_gene_ensembl     zfin_id_id",
        sp_source_data={},
    ),
    SpeciesRow(
        sp_id=5,
        sp_name="Drosophila melanogaster",
        sp_tax_id=7227,
        sp_ref_gdb_id=14,
        sp_date="2011-03-14",
        sp_biomart_info="dmelanogaster_gene_ensembl      flybase_gene_id",
        sp_source_data={},
    ),
    SpeciesRow(
        sp_id=6,
        sp_name="Macaca mulatta",
        sp_tax_id=9544,
        sp_ref_gdb_id=1,
        sp_date="2011-03-14",
        sp_biomart_info="mmulatta_gene_ensembl   ensembl_gene_id",
        sp_source_data={},
    ),
    SpeciesRow(
        sp_id=8,
        sp_name="Caenorhabditis elegans",
        sp_tax_id=6239,
        sp_ref_gdb_id=15,
        sp_date="2011-03-14",
        sp_biomart_info="celegans_gene_ensembl   wormbase_gene",
        sp_source_data={},
    ),
    SpeciesRow(
        sp_id=9,
        sp_name="Saccharomyces cerevisiae",
        sp_tax_id=559292,
        sp_ref_gdb_id=16,
        sp_date="2011-03-14",
        sp_biomart_info="scerevisiae_gene_ensembl        entrezgene_id",
        sp_source_data={},
    ),
    SpeciesRow(
        sp_id=10,
        sp_name="Gallus gallus",
        sp_tax_id=9031,
        sp_ref_gdb_id=20,
        sp_date="2011-03-14",
        sp_biomart_info="ggallus_gene_ensembl    entrezgene_id",
        sp_source_data={},
    ),
    SpeciesRow(
        sp_id=11,
        sp_name="Canis familiaris",
        sp_tax_id=9615,
        sp_ref_gdb_id=1,
        sp_date="2011-03-14",
        sp_biomart_info="cfamiliaris_gene_ensembl        entrezgene_id",
        sp_source_data={},
    ),
]



"""Parse various score types from string inputs.

This module provides a set of functions for parsing various types of
numerical scores from string inputs and converting them into GenesetScoreType
objects. It also provides functionality for extracting numerical values from
strings and ensuring the correct count of these values.

Parsing Functions:
- parse_score: Parse a score input string into a GenesetScoreType object.
- parse_binary: Convert a binary score input string into a GenesetScoreType object.
- parse_pvalue: Convert a p-value score input string into a GenesetScoreType object.
- parse_qvalue: Convert a q-value score input string into a GenesetScoreType object.
- parse_correlation: Convert a correlation score input string into a GenesetScoreType
object.
- parse_effect: Convert an effect score input string into a GenesetScoreType object.

Number Extraction Functions:
- extract_numerical_values: Extract all numerical values from a given string.
- extract_single_numerical_value: Extract a single numerical value from a given string.
- extract_one_or_two_numerical_values: Extract either one or two numerical values from a
given string.

Exceptions:
- InvalidScoreThresholdException: Raised when the number of numerical values extracted
  from a string does not meet the expected count.
"""

import re
from typing import List

from geneweaver.core.parse.exceptions import InvalidScoreThresholdError
from geneweaver.core.schema.score import GenesetScoreType, ScoreType


def parse_binary(score_input: str) -> GenesetScoreType:
    """Convert a binary score input string into a GenesetScoreType object.

    This function expects a binary score as input, and will convert it into a
    GenesetScoreType object with a score type of 'BINARY' and a threshold of 1.

    :param score_input: The binary score input as a string (e.g. 'binary'). This input
    will be ignored.

    :returns: An object representing the binary score type and its threshold.
    """
    return GenesetScoreType(score_type=ScoreType.BINARY, threshold=1)


def parse_pvalue(score_input: str) -> GenesetScoreType:
    """Convert a p-value score input string into a GenesetScoreType object.

    This function takes a p-value score input string, parses the p-value using the
    `extract_single_numeric_value` function, and returns a GenesetScoreType object with
    the score type set to 'P_VALUE' and the threshold set to the parsed p-value.

    :param score_input: The p-value score input as a string.

    :returns: An object representing the p-value score type and its threshold.
    """
    threshold = extract_single_numeric_value(score_input)
    return GenesetScoreType(score_type=ScoreType.P_VALUE, threshold=threshold)


def parse_qvalue(score_input: str) -> GenesetScoreType:
    """Convert a q-value score input string into a GenesetScoreType object.

    This function takes a q-value score input string, parses the q-value using the
    `extract_single_numeric_value` function, and returns a GenesetScoreType object with
    the score type set to 'Q_VALUE' and the threshold set to the parsed q-value.

    :param score_input: The q-value score input as a string.

    :returns: An object representing the q-value score type and its threshold.
    """
    threshold = extract_single_numeric_value(score_input)
    return GenesetScoreType(score_type=ScoreType.Q_VALUE, threshold=threshold)


def parse_correlation(score_input: str) -> GenesetScoreType:
    """Convert a correlation score input string into a GenesetScoreType object.

    This function takes a correlation score input string, parses the thresholds using
    the `extract_one_or_two_numeric_values` function, and returns a GenesetScoreType
    object with the score type set to 'CORRELATION', the low threshold set to the first
    parsed value, and the high threshold set to the second parsed value.

    :param score_input: The correlation score input as a string.

    :returns: An object representing the correlation score type and its thresholds.
    """
    thresholds = extract_one_or_two_numeric_values(score_input)
    return GenesetScoreType(
        score_type=ScoreType.CORRELATION,
        threshold_low=thresholds[0],
        threshold=thresholds[1],
    )


def parse_effect(score_input: str) -> GenesetScoreType:
    """Convert an effect score input string into a GenesetScoreType object.

    This function takes an effect score input string, parses the thresholds using the
    `extract_one_or_two_numeric_values` function, and returns a GenesetScoreType object
    with the score type set to 'EFFECT', the low threshold set to the first parsed
    value, and the high threshold set to the second parsed value.

    :param score_input: The effect score input as a string.

    :returns: An object representing the effect score type and its thresholds.
    """
    thresholds = extract_one_or_two_numeric_values(score_input)
    return GenesetScoreType(
        score_type=ScoreType.EFFECT,
        threshold_low=thresholds[0],
        threshold=thresholds[1],
    )


# This dictionary maps score types to their respective parsing functions. It is used by
# the `parse_score` function to determine which parsing function to use for a given
# score type.
SCORE_PARSER_MAP = {
    "binary": parse_binary,
    "p-value": parse_pvalue,
    "q-value": parse_qvalue,
    "correlation": parse_correlation,
    "effect": parse_effect,
}


def parse_score(score_input: str) -> GenesetScoreType:
    """Parse a score input string into a GenesetScoreType object.

    This function takes a score input string, converts it to lowercase, and
    checks it against a mapping of score types to their respective parsers.
    If a match is found, it uses the corresponding parser to convert the input
    string into a GenesetScoreType object.

    Example valid `score_input` score type strings:
    'Binary'
    'P-Value < 0.05'
    'Q-Value < 0.05'
    '0.40 < Correlation < 0.90'
    '6.0 < Effect < 22.50'

    If no matching score type is found in the mapping, the function raises an
    InvalidScoreThresholdException.

    :param score_input: The score input as a string.

    :returns: An object representing the parsed score type and its thresholds.

    :raises InvalidScoreThresholdException: If no matching score type can be found in
    the mapping.
    """
    score_type = score_input.lower()
    for score_key, score_parser in SCORE_PARSER_MAP.items():
        if score_key in score_type:
            print(score_key, score_type)
            return score_parser(score_input)

    raise InvalidScoreThresholdError()


def extract_numeric_values(score_input: str) -> List[str]:
    """Extract all numerical values from a given string.

    This function uses regular expressions to find and return all numeric
    values in the input string. The numeric values can be integers or floats, positive
    or negative. The values can be separated by whitespace or other characters.

    :param score_input: The input string to search for numeric values.

    :returns: A list of all numeric values found in the input string.
    """
    # This regex uses negative lookbehind to ensure that a minus sign is not preceded
    # by a digit, which would indicate that it is a subtraction operator or a dash
    # separator. The simpler version of this regex, without the negative lookbehind,
    # would look like `(-?[0-9.]+)`, but would incorrectly match the dash in the
    # in a range as a negative sign (e.g. '4-5' would match as ['4', '-5'] instead of
    # ['4', '5']).
    return re.findall(r"((?<![0-9])-?[0-9.]+)", score_input)


def extract_single_numeric_value(score_input: str) -> float:
    """Extract a single numerical value from a given string.

    This function uses the `extract_numerical_values` function to find all numeric
    values in the input string, and ensures that exactly one value is present. If not,
    it raises an `InvalidScoreThresholdException`.

    :param score_input: The input string to search for a single numeric value.

    :returns: The single numeric value found in the input string.

    :raises InvalidScoreThresholdException: If there is not exactly one numeric value in
    the input string.
    """
    values = extract_numeric_values(score_input)
    if len(values) != 1:
        raise InvalidScoreThresholdError()

    return float(values[0])


def extract_one_or_two_numeric_values(score_input: str) -> list:
    """Extract either one or two numerical values from a given string.

    This function uses the `extract_numerical_values` function to find all numeric
    values in the input string, and ensures that either one or two values are present.
    If not, it raises an `InvalidScoreThresholdException`. If only one value is found,
    the function returns a list with `None` as the first element and the found value as
    the second element.

    :param score_input: The input string to search for one or two numeric values.

    :returns: A list of one or two numeric values found in the input string.

    :raises InvalidScoreThresholdException: If there are not either one or two numeric
    values in the input string.
    """
    values = extract_numeric_values(score_input)
    if len(values) not in (1, 2):
        raise InvalidScoreThresholdError()

    if len(values) == 1:
        values = [None, float(values[0])]
    else:
        values = [float(v) for v in values]

    return values



"""Functions for parsing numpy objects.

The top level functions are (in-order):
- ndarray_to_gene_values
- ndarray_to_gene_values_by_idx
- ndarray_to_gene_values_named

If you know that your numpy array has the gene symbol in the first column and the
gene value in the second column, you can use ndarray_to_gene_values_by_idx.
"""
from typing import List, Tuple

import numpy as np
from geneweaver.core.schema.batch import GenesetValueInput


def ndarray_to_gene_values(geneset_array: np.ndarray) -> List[GenesetValueInput]:
    """Convert a numpy array to a list of GeneValueInput objects.

    This function will try to map via labels first, then by index.

    :param geneset_array: The numpy array to convert.
    :raises ValueError: If there's a problem mapping the numpy array.
    :return: A set of the values in the numpy array.
    """
    check_input_shape_and_type(geneset_array)
    try:
        return ndarray_to_gene_values_named(geneset_array)
    except ValueError:
        return ndarray_to_gene_values_by_idx(geneset_array)


def ndarray_to_gene_values_by_idx(geneset_array: np.ndarray) -> List[GenesetValueInput]:
    """Convert a numpy array to a list of GeneValueInput objects.

    The numpy array must be a 2-dimensional array with 2 columns. The first column
    will be mapped to the GenesetValueInput.symbol attribute, and the second column will
    be mapped to the GenesetValueInput.value attribute.

    :param geneset_array: The numpy array to convert.
    :return: A set of the values in the numpy array.
    """
    try:
        return [GenesetValueInput(symbol=row[0], value=row[1]) for row in geneset_array]
    except IndexError as e:
        raise ValueError("Input must be a 2-dimensional array") from e


def ndarray_to_gene_values_named(geneset_array: np.ndarray) -> List[GenesetValueInput]:
    """Convert a numpy array to a list of GeneValueInput objects using labels.

    :param geneset_array: The numpy array to convert.
    :raises ValueError: If there's a problem mapping the numpy array using labels.
    :return: A set of the values in the numpy array.
    """
    if not ndarray_has_gene_labels(geneset_array):
        raise ValueError("Numpy array does not have gene labels")

    symbol_key, value_key = map_ndarray_labels_to_gene_value_attr(geneset_array)

    return [
        GenesetValueInput(symbol=row[symbol_key], value=row[value_key])
        for row in geneset_array
    ]


def check_input_shape_and_type(input_array: np.ndarray) -> None:
    """Check the shape and type of the input array.

    :raises ValueError: If the input array is not a numpy array or if it is not

    :param input_array: The array to check.
    """
    if not isinstance(input_array, np.ndarray):
        raise ValueError("Input must be a numpy array")

    if len(input_array.shape) != 2 and not input_array.dtype.names:
        raise ValueError("Input must be a 2-dimensional array, or a structured array")


def ndarray_has_gene_labels(geneset_array: np.ndarray) -> bool:
    """Check if the numpy array has gene labels.

    This function is used to check if a numpy array is structured (has names),
    and if it does, that it has at least 2 names.

    :param geneset_array: The numpy array to check.
    :return: True if the numpy array has gene labels, False otherwise.
    """
    if geneset_array.dtype.names is None:
        return False

    if len(geneset_array.dtype.names) < 2:
        return False

    return True


SYMBOL_KEYS = ("Symbol", "GeneID", "Gene_ID", "Gene ID")
VALUE_KEYS = ("Value", "Score", "PValue", "QValue", "Effect", "Correlation")


def map_ndarray_labels_to_gene_value_attr(geneset_array: np.ndarray) -> Tuple[str, str]:
    """Map numpy array labels to GenesetValueInput attributes.

    This function is used to check if a numpy array is structured (has names),
    and if it does, it checks if the names can be mapped to a GenesetValueInput object.

    In order, it will map ("Symbol", "GeneID", "Gene_ID", "Gene ID") to the
    GenesetValueInput symbol attribute.

    In order, it will map
    ("Value", "Score", "PValue", "QValue", "Effect", "Correlation") to the
    GenesetValueInput value attribute.

    If it cannot map to both symbol and value, it will raise an error.

    :param geneset_array: The numpy array to map.

    :raises ValueError: If the numpy array labels cannot be mapped to GenesetValueInput
    attributes.
    :return: A tuple, where the first value is the key mapping to the symbol attribute
    and the second value is the key mapping to the value attribute.
    """
    symbol_key, value_key = None, None
    for key in SYMBOL_KEYS:
        if key in geneset_array.dtype.names:
            symbol_key = key
            break

    for key in VALUE_KEYS:
        if key in geneset_array.dtype.names:
            value_key = key
            break

    if symbol_key is None or value_key is None:
        raise ValueError(
            "Could not map numpy array labels to GenesetValueInput attributes"
        )

    return symbol_key, value_key



"""Python Package main config file.

For more options refer to pydantic base-settings docs:
https://pydantic-docs.helpmanual.io/usage/settings/
"""
from pydantic import BaseSettings


class ExternalServiceSettings(BaseSettings):
    """External Service Config and Settings Configuration."""

    PUBMED_XLM_SVC_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id={0}&retmode=xml"


class CoreSettings(BaseSettings):
    """Root Config and Settings Configuration."""

    PROJECT_NAME = "jax-geneweaver-core"
    VERSION = "0.0.2"
    LOG_LEVEL: str = "INFO"
    SERVICE_URLS: ExternalServiceSettings = ExternalServiceSettings()

    class Config:
        """Pydantic Config class."""

        env_prefix = "GW_"



"""Pydantic schema for defining score types."""
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class ScoreType(Enum):
    """Enum for defining score types."""

    P_VALUE = 1
    Q_VALUE = 2
    BINARY = 3
    CORRELATION = 4
    EFFECT = 5


class GenesetScoreType(BaseModel):
    """Pydantic schema for defining score types."""

    score_type: ScoreType
    threshold_low: Optional[float] = None
    threshold: float = 0.05

    def __str__(self) -> str:
        name = self.score_type.name.title().replace("_", "-")
        if self.threshold_low:
            return f"{self.threshold_low} < {name} < {self.threshold}"
        return f"{name} < {self.threshold}"



"""Schema for tools and tool framework."""
from typing import List

from pydantic import BaseModel

from .geneset import GenesetGenes


class Tool(BaseModel):
    """Tool schema."""

    name: str
    description: str
    version: str
    url: str


class ToolInput(BaseModel):
    """Tool input schema."""

    Metadata: dict
    GeneSets: List[GenesetGenes]


class ToolOutput:
    """Tool output schema."""

    Metadata: dict
    Images: List[str]
    GeneSets: List[GenesetGenes]



"""Custom exceptions for the GeneWeaver project."""


class GeneweaverError(Exception):
    """Base class for all Geneweaver exceptions."""

    pass


class ExternalAPIError(GeneweaverError):
    """Base class for all external API exceptions."""

    pass



"""Schemas for geneset."""
import datetime
from typing import List, Optional

from geneweaver.core.enum import GenesetAccess, GenesetScoreType
from geneweaver.core.schema.gene import GeneValue
from pydantic import BaseModel, Field


class Geneset(BaseModel):
    """Geneset schema."""

    name: str
    abbreviation: str
    description: str
    count: int
    threshold_type: int
    threshold: str
    gene_id_type: int
    created: datetime.date
    admin_flag: str
    updated: datetime.datetime
    status: str
    gsv_qual: str
    attribution: int
    is_edgelist: bool


class GenesetGenes(BaseModel):
    """Geneset genes schema."""

    genes: List[GeneValue]


class GenesetUpload(BaseModel):
    """Geneset upload schema."""

    name: str
    label: str
    score_type: GenesetScoreType
    description: str
    pubmed_id: Optional[str]
    access: GenesetAccess
    groups: List[str] = []
    species: str
    gene_identifier: str
    gene_list: List[GeneValue]


class BatchUpload(BaseModel):
    """Batch upload schema."""

    batch_file: str
    curation_group: List[str]


class GenesetInfo(BaseModel):
    """Geneset info schema."""

    id: int  # noqa: A003
    page_views: int
    referers: List[str]
    analyses: List[str]
    resource_id: int
    last_sim: str
    last_ann: str
    jac_started: str
    jac_completed: str


class SimilarGeneset(Geneset):
    """Schema for similar geneset relation."""

    jax_value: float
    gic_value: float


class GenesetRow(BaseModel):
    """Geneset schema for database row."""

    gs_id: int
    usr_id: int
    file_id: int
    gs_name: str
    gs_abbreviation: str
    pub_id: int
    res_id: int
    cur_id: int
    gs_description: str
    sp_id: int
    gs_count: int
    gs_threshold_type: int
    gs_threshold: str
    gs_groups: str
    gs_attribution_old: str
    gs_uri: str
    gs_gene_id_type: int
    gs_created: datetime.date
    admin_flag: str
    gs_updated: datetime.datetime
    gs_status: str
    gsv_qual: str
    _comments_author: str
    _comments_curator: str
    gs_attribution: int
    gs_is_edgelist: bool



"""Utility functions for use across the geneweaver core package."""
from typing import Any


def add_to_dict_if_not_none(d: dict, key: str, value: Any) -> None:  # noqa: ANN401
    """Add a value to a dictionary only if the value is not None.

    :param dict d: The dictionary to which the value should be added.
    :param str key: The key under which the value should be added.
    :param Any value: The value to be added. If this is None, the function does nothing.
    """
    if value is not None:
        d[key] = value



"""Utility functions to parse CSV documents."""
import csv
import itertools
from typing import Dict, List, TextIO, Tuple, Union

from geneweaver.core.parse.exceptions import EmptyFileError
from geneweaver.core.types import StringOrPath


def find_header(
    file_path: StringOrPath, max_rows_to_check: int = 5
) -> Tuple[bool, int]:
    """Determine if a CSV (.csv) file has a header row.

    This function will check up to 'max_rows_to_check' rows from the top to find a
    header row.

    This function assumes that a row can be interpreted as a header if none of its
    values are numeric, if it has more than one column, and if the following row has the
    same number of columns that it does.

    :param file_path: Path to the CSV file.
    :param max_rows_to_check: The number of rows to check from the top to find a header
    row.

    :returns: First index - True if a header row is found, False otherwise.
              Second index - The index of the header row if found, otherwise -1.
    """
    with open(file_path, "r") as f:
        reader = csv.reader(f)
        for i, row in enumerate(reader):
            if i >= max_rows_to_check:
                break
            if (
                len(row) > 1
                and all(len(item) > 0 for item in row)
                and all(not item.replace(".", "", 1).isdigit() for item in row)
            ):
                next_row = next(reader, None)
                if next_row and len(row) == len(next_row):
                    return True, i
    return False, -1


def has_header(file_path: StringOrPath, max_rows_to_check: int = 5) -> bool:
    """Summary function to return true/false for if a header is found by `find_header`.

    :param file_path: Path to the CSV file.
    max_rows_to_check (int): The number of rows to check from the top to find a header
    row.

    :returns:  True if a header row is found, False otherwise.
    """
    return find_header(file_path, max_rows_to_check)[0]


def get_headers(file_path: StringOrPath) -> Tuple[List[str], int]:
    """Read the first row from a CSV file and return it as a list of strings.

    Assumes that the first row of the CSV file is the header row.

    :param file_path: Path to the CSV file.

    :returns: List of column names from the CSV file, and the index of the
    header row.
    """
    headers = []
    file_has_header, header_idx = find_header(file_path)
    if file_has_header:
        headers = read_row(file_path, header_idx)
    return headers, header_idx


def read_row(file_path: StringOrPath, row_idx: int = 0) -> List[str]:
    """Get the contents of a row from a CSV file.

    :param file_path: The file path to the CSV file.
    :param row_idx: The index of the row from which to read the headers. Defaults to 0.

    :returns: The contents of a row

    :raises ValueError: If the file is empty or does not contain enough rows.
    """
    row = []
    with open(file_path, newline="") as f:
        reader = csv.reader(f)
        for i, r in enumerate(reader):
            if i == row_idx:
                row = r
                break
        else:
            raise ValueError(f"File does not contain a row at index {row_idx}")
        return row


def get_csv_dict_reader(file: TextIO, start_row: int) -> csv.DictReader:
    """Get a csv.DictReader for the given file, starting from the specified row.

    :param file: The opened file object.
    :param start_row: The row number to start reading from (0-indexed). This row will be
    used as the header.

    :returns: A DictReader instance for the remaining rows in the file, using the
    specified row as the header.
    """
    reader = csv.reader(file)
    header = None
    for _ in range(start_row + 1):  # read up to and including the start_row
        header = next(reader, None)
    if header is None:
        raise ValueError("start_row was larger than the number of rows in the file")

    return csv.DictReader(file, fieldnames=header)


def read_to_dict(
    file_path: StringOrPath, start_row: int = 0
) -> List[Dict[str, Union[str, int]]]:
    """Read a CSV file and return its contents as a list of dictionaries.

    Each dictionary in the list corresponds to a row in the CSV file, and the keys
    are the column names from the CSV file.

    :param file_path: Path to the CSV file.
    :param start_row: The row number to start reading from (0-indexed).
                               This row will be used as the header.
                               Defaults to 0.

    :returns: List of dictionaries representing the CSV file.
    """
    with open(file_path, mode="r") as infile:
        dict_reader = get_csv_dict_reader(infile, start_row)
        data = [dict(row) for row in dict_reader]
    return data


def read_to_dict_n_rows(
    file_path: StringOrPath, n: int, start_row: int = 0
) -> List[Dict[str, str]]:
    """Parse n lines of a CSV file into a dictionary.

    Parse a CSV file into a dictionary, with keys as column names and values as column
    values. This function will only return the first n rows of data.

    :param file_path: The file path to the CSV file.
    :param n: The number of rows of data to return.
    :param start_row: The row number to start reading from (0-indexed).
                               This row will be used as the header.
                               Defaults to 0.

    :returns: A list of dictionaries, where each dictionary represents a row from the
    CSV file.
    """
    with open(file_path, mode="r") as infile:
        dict_reader = get_csv_dict_reader(infile, start_row)
        data = list(itertools.islice(dict_reader, n))
        if len(data) == 0:
            raise EmptyFileError(
                file_path,
                f"Selected start row ({start_row}) and n ({n}) yielded no results.",
            )
    return data



"""Ontology schema."""
from typing import List

from pydantic import BaseModel


class Ontology(BaseModel):
    """Ontology schema."""

    ontology_id: int
    reference_id: int
    name: str
    description: str
    children: List["Ontology"]
    parents: List["Ontology"]
    ontdb_id: int
    ro_ont_id: int


class OntologyDB(BaseModel):
    """Ontology database schema."""

    ontology_db_id: int
    name: str
    prefix: str
    ncbo_id: str
    date: str
    linkout_url: str
    ncbo_vid: str
